from PySide6.QtCore import Qt
from PySide6.QtCore import Slot
from PySide6.QtCore import QThread
from PySide6.QtCore import Signal

from PySide6.QtGui import QAction
from PySide6.QtGui import QIcon

from PySide6.QtWidgets import QMainWindow
from PySide6.QtWidgets import QToolBar
from PySide6.QtWidgets import QDockWidget
from PySide6.QtWidgets import QProgressBar

from pathlib import Path

from mimetica import conf
from mimetica import Canvas
from mimetica import Dock
from mimetica import SplitView
from mimetica import Stack


class Tab(QMainWindow):
    load_stack = Signal()

    def __init__(
        self,
        paths: list[Path],
    ):
        QMainWindow.__init__(self)

        self.paths = paths

        # The display canvas
        # ==================================================
        self.canvas = Canvas()

        # Splitview widget
        # ==================================================
        self.splitview = SplitView(self.canvas)
        self.splitview.setSizes((2, 1))
        self.setCentralWidget(self.splitview)

        # Dock
        # ==================================================
        self.dock = Dock(features=QDockWidget.DockWidgetFeature.DockWidgetClosable)
        self.addDockWidget(Qt.DockWidgetArea.LeftDockWidgetArea, self.dock)

        # Toolbar
        # ==================================================
        self.toolbar = QToolBar(floatable=False, movable=False)
        self.addToolBar(Qt.ToolBarArea.TopToolBarArea, self.toolbar)

        # Thread for preventing the GUI from blocking
        # ==================================================
        self.worker = QThread()
        self.worker.start()

        # The layer stack
        # ==================================================
        # logger.info(f'Loading stack | TID: {threading.get_ident()}')
        self.stack = Stack(paths, conf.show_inactive_plots)
        self.load_stack.connect(self.stack.process)
        self.stack.set_canvas.connect(self.set_canvas)
        self.stack.abort.connect(self._abort)
        self.stack.moveToThread(self.worker)

        # Status bar
        # ==================================================
        self.status_bar = self.statusBar()
        self.progress_bar = QProgressBar()
        self.status_bar.addPermanentWidget(self.progress_bar)
        self.progress_bar.setGeometry(30, 40, 200, 25)
        self.progress_bar.setValue(0)
        self.progress_bar.hide()
        self.setup_statusbar()

        # Slots & signals
        # ==================================================
        self.stack.update_progress.connect(self._update_progress_bar)
        self.dock.sig_show_inactive_plots.connect(self.splitview._show_inactive_plots)
        self.dock.sig_set_active_plot_colour.connect(
            self.splitview._set_active_plot_colour
        )
        self.dock.sig_set_inactive_plot_colour.connect(
            self.splitview._set_inactive_plot_colour
        )
        self.dock.sig_set_stack_contour_colour.connect(
            self.canvas._set_stack_contour_colour
        )
        self.dock.sig_set_slice_contour_colour.connect(
            self.canvas._set_slice_contour_colour
        )
        self.dock.sig_show_stack.connect(self.canvas._set_show_stack)
        self.dock.sig_set_radial_segments.connect(self.canvas._slot_compute_radial_profile)
        self.dock.sig_set_phase_segments.connect(self.canvas._slot_compute_phase_profile)

        # Load the tab
        # ==================================================
        self._load_tab()

    def setup_statusbar(self):

        pass

    def _update_progress_bar(self, path: Path):
        self.status_bar.showMessage(f"Processing {path}")
        self.progress_bar.setValue(self.progress_bar.value() + 1)

    def setup_toolbar(self):

        # Dock widget toggle
        act_dock = QAction(
            QIcon.fromTheme("document-properties"), "Show docking panel", self.toolbar
        )
        act_dock.triggered.connect(self._toggle_dock)
        self.toolbar.addAction(act_dock)

        # Scale image button
        act_scale_image = QAction(
            QIcon.fromTheme("zoom-fit-best"), "Scale image to fit", self.toolbar
        )
        act_scale_image.triggered.connect(self.canvas._reset_zoom)
        self.toolbar.addAction(act_scale_image)

        # # Find centre
        # act_find_centre = QAction(QIcon.fromTheme("tools-media-optical-format"), "Reset centre", self.toolbar)
        # act_find_centre.triggered.connect(lambda: self.canvas._reset_centre())
        # self.toolbar.addAction(act_find_centre)

        # # Radial profile button
        # act_radial_profile = QAction(QIcon.fromTheme("object-rotate-left"), "Plot radial profile", self.toolbar)
        # act_radial_profile.triggered.connect(self.canvas.layer.compute_radial_profile)
        # self.toolbar.addAction(act_radial_profile)

        # # Radial plot button
        # act_plot = QAction(QIcon.fromTheme("list-add"), "Plot radial profile", self.toolbar)
        # act_plot.triggered.connect(lambda: self.splitview.plot(self.stack.current_layer))
        # self.toolbar.addAction(act_plot)

    @Slot()
    def _toggle_dock(self):
        if self.dock.isVisible():
            self.dock.hide()
        else:
            self.dock.show()

    @Slot()
    def _set_threshold(
        self,
        update: bool = True,
    ):
        self.dock._slot_show_inactive_plots()
        self.stack._update_threshold(self.dock.show_inactive_plots)

    def _load_tab(self):
        self.status_bar.showMessage(f"Loading stack from {self.paths[0].parent}...")
        self.progress_bar.show()
        self.load_stack.emit()

    def _abort(self):
        self.worker.quit()

    @Slot()
    def set_canvas(self):

        self.status_bar.showMessage(f"Setting up canvas...")
        self.progress_bar.setValue(0)
        self.progress_bar.hide()
        self.worker.quit()
        self.canvas.set_stack(self.stack, auto_range=True)
        self.status_bar.clearMessage()
        self.setup_toolbar()



from PySide6 import QtGui
from PySide6.QtGui import QPainter

import pyqtgraph as pg
from pyqtgraph import functions as pgfn


class RadialLine(pg.ROI):
    """
    ROI subclass with two freely-moving handles defining a line.
    """

    def __init__(
        self,
        positions=(None, None),
        pos=None,
        **args,
    ):
        if pos is None:
            pos = [0, 0]

        self.endpoints = [pg.Point(p) for p in positions]
        pg.ROI.__init__(self, pos, [1, 1], **args)
        if len(positions) > 2:
            raise Exception(
                "LineROI must be defined by exactly 2 positions. For more points, use PolyLineROI."
            )

    def paint(self, p: QPainter, *args):
        p.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        p.setPen(self.currentPen)
        p.drawLine(self.endpoints[0], self.endpoints[1])

    def listPoints(self):
        return self.endpoints

    def getState(self):
        state = pg.ROI.getState(self)
        state["points"] = self.endpoints
        return state

    def saveState(self):
        state = pg.ROI.saveState(self)
        state["points"] = [tuple(p) for p in self.endpoints]
        return state

    def setState(self, state):
        pg.ROI.setState(self, state)
        p1 = [
            state["points"][0][0] + state["pos"][0],
            state["points"][0][1] + state["pos"][1],
        ]
        p2 = [
            state["points"][1][0] + state["pos"][0],
            state["points"][1][1] + state["pos"][1],
        ]
        self.movePoint(self.endpoints[0], p1, finish=False)
        self.movePoint(self.endpoints[1], p2)

    def boundingRect(self):
        return self.shape().boundingRect()

    def shape(self):
        p = QtGui.QPainterPath()

        ep1 = self.endpoints[0]
        ep2 = self.endpoints[1]
        dh = ep2 - ep1
        if dh.length() == 0:
            return p
        pxv = self.pixelVectors(dh)[1]
        if pxv is None:
            return p

        pxv *= 4

        p.moveTo(ep1 + pxv)
        p.lineTo(ep2 + pxv)

        return p

    def set_end(
        self,
        pos: tuple[float, float],
    ):
        self.endpoints[-1] = pg.Point(pos)

    def getArrayRegion(
        self, data, img, axes=(0, 1), order=1, returnMappedCoords=False, **kwds
    ):
        """
        Use the position of this ROI relative to an imageItem to pull a slice
        from an array.

        Since this pulls 1D data from a 2D coordinate system, the return value
        will have ndim = data.ndim-1

        See :meth:`~pyqtgraph.ROI.getArrayRegion` for a description of the
        arguments.
        """
        imgPts = [self.mapToItem(img, ep) for ep in self.endpoints]

        d = pg.Point(imgPts[1] - imgPts[0])
        o = pg.Point(imgPts[0])
        rgn = pgfn.affineSlice(
            data,
            shape=(int(d.length()),),
            vectors=[pg.Point(d.norm())],
            origin=o,
            axes=axes,
            order=order,
            returnCoords=returnMappedCoords,
            **kwds,
        )

        return rgn



from pathlib import Path

import numpy as np

import shapely as shp

from concurrent.futures import ProcessPoolExecutor

from PySide6.QtCore import Slot
from PySide6.QtCore import Signal
from PySide6.QtCore import QObject

from mimetica import Layer
from mimetica import utils
from mimetica import logger


class Stack(QObject):
    update_progress = Signal(Path)
    set_canvas = Signal()
    abort = Signal()

    @staticmethod
    def make_layer(
        path: str | Path,
    ):
        """
        Create a layer for the given image.

        Args:
            path: Path to the image file.

        Returns:
            A layer instance.
        """
        return Layer(path)

    def __init__(
        self,
        paths: list[Path],
        threshold: int = 70,
        *args,
        **kwargs,
    ):
        """
        Create a stack of images.

        Args:
            paths: A list of image paths.
            threshold: Binarisation threshold (only for RGB or greyscale images; currently unused)
        """
        super().__init__(*args, **kwargs)

        # Save the parameters
        # ==================================================
        self.paths = sorted(paths)
        self.threshold = threshold

        # Other attributes
        # ==================================================
        self.layers = []
        self.active_layer = 0

        # Create a merged stack
        # ==================================================
        self.merged: np.ndarray = None

    def _set_active_layer(
        self,
        index: int = 0,
    ):
        """
        Set the layer to the given index.

        Args:
            index: Layer index.
        """
        self.active_layer = index

    @Slot(int)
    def _slot_compute_radial_profile(
        self,
        segments: int,
    ):
        '''
        Update the radial profile with a new number of segments.

        Args:
            segments: Number of radial segments.
        '''

        for layer in self.layers:
            layer.compute_radial_profile()
        self.plot.emit()

    @Slot(int)
    def _slot_compute_phase_profile(
        self,
        segments: int,
    ):
        '''
        Update the phase profile with a new number of segments.

        Args:
            segments: Number of phase segments.
        '''
        for layer in self.layers:
            layer.compute_phase_profile()

        self.plot.emit()

    @Slot()
    def process(self):
        """
        Process a stack of images.
        """
        logger.info(f"Loading stack...")

        # Layer factory
        # ==================================================
        with ProcessPoolExecutor() as executor:
            for layer in executor.map(Stack.make_layer, self.paths):
                self.layers.append(layer)
                self.update_progress.emit(layer.path)

        self._set_active_layer()

        # Calibrate the stack based on all the images
        # ==================================================
        for layer in self.layers:

            if self.merged is None:
                self.merged = layer.image.copy().astype(np.float32)
            else:
                self.merged += layer.image

        # Scale the merged stack
        # ==================================================
        minval = self.merged.min()
        maxval = self.merged.max()
        self.merged = (255 * (self.merged - minval) / (maxval - minval)).astype(
            np.uint8
        )

        # Set the stack on the canvas
        # ==================================================
        self.set_canvas.emit()



import pyqtgraph as pg


class Contour(pg.EllipseROI):
    """
    Circular contour subclass.
    """

    def __init__(self, pos, radius, **args):

        args.setdefault("movable", False)
        args.setdefault("resizable", False)
        args.setdefault("removable", False)
        args.setdefault("rotatable", False)

        pg.EllipseROI.__init__(
            self,
            pos,
            1,
            aspectLocked=True,
            **args,
        )

        for h in self.getHandles():
            self.removeHandle(h)

        self.setSize(2 * radius, (0.5, 0.5), update=True, finish=True)



from PySide6.QtCore import Qt
from PySide6.QtCore import Slot
from PySide6.QtCore import Signal
from PySide6.QtCore import QEvent
from PySide6.QtCore import QObject

from PySide6.QtGui import QKeyEvent
from PySide6.QtGui import QMouseEvent
from PySide6.QtGui import QEnterEvent

from PySide6.QtWidgets import QWidget
from PySide6.QtWidgets import QGridLayout
from PySide6.QtWidgets import QHBoxLayout
from PySide6.QtWidgets import QScrollArea
from PySide6.QtWidgets import QSizePolicy

import numpy as np
import skimage as ski

import pyqtgraph as pg
from pyqtgraph.GraphicsScene.mouseEvents import MouseClickEvent
from pyqtgraph import SignalProxy

from mimetica import Thumbnail
from mimetica import ImageView
from mimetica import Layer
from mimetica import Stack
from mimetica import conf
from mimetica.gui.display.roi.contour import Contour
from mimetica.gui.display.roi.target import Target


class Canvas(QWidget):
    plot = Signal()
    highlight_plot = Signal(int)
    update_radial_plot = Signal(float)
    update_phase_plot = Signal(float)

    class EventHandler(QObject):

        ctrl_signal = Signal(bool)
        focus_signal = Signal()

        def __init__(self, wh, *args, **kwargs):

            super().__init__(*args, **kwargs)
            self.wh = wh

        def eventFilter(self, obj: QObject, event: QEvent):
            if obj is self.wh:
                if isinstance(event, QEnterEvent):
                    self.focus_signal.emit()
                elif isinstance(event, QKeyEvent) and event.key() == Qt.Key.Key_Control:
                    if event.type() == QKeyEvent.Type.KeyPress:
                        self.ctrl_signal.emit(True)
                    elif event.type() == QKeyEvent.Type.KeyRelease:
                        self.ctrl_signal.emit(False)
            return QObject().eventFilter(obj, event)

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        # Image viewport
        # ==================================================
        self.iv = ImageView(self)

        # Current stack
        # ==================================================
        self.stack = None

        # Arrays used for displaying parts of the layer
        # ==================================================
        self.image = None
        self.centre = None
        self.stack_mbc = None
        self.layer_mbcs = []
        self.phase_resolution = 1

        # Layout grid
        # ==================================================
        self.grid = QGridLayout(self)
        self.grid.addWidget(self.iv, 0, 0, 1, 2)

        # Thumbnails
        # ==================================================
        self.thumbnails = []
        self.tb_widget = QWidget()
        self.tb_scroll_area = QScrollArea(self)
        self.tb_layout = QHBoxLayout()
        self.tb_layout.setContentsMargins(0, 0, 0, 0)
        self.tb_widget.setLayout(self.tb_layout)

        # Layer highlighters
        # ==================================================
        self.slice_contour_pen = pg.mkPen(color=conf.slice_contour_colour, width=1)
        self.slice_contour = None
        self.slice_centre_pen = pg.mkPen(
            color=conf.slice_contour_colour, width=1
        )  # TODO: Separate conf entry
        self.slice_centre = None

        # Stack highlighters
        # ==================================================
        self.stack_contour_pen = pg.mkPen(color=conf.stack_contour_colour, width=1)
        self.stack_contour = None

        self.tb_scroll_area.setFixedHeight(110)
        self.tb_scroll_area.setHorizontalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAsNeeded
        )
        self.tb_scroll_area.setVerticalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAlwaysOff
        )
        self.tb_scroll_area.setSizePolicy(
            QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum
        )
        self.tb_scroll_area.setWidgetResizable(True)
        self.tb_scroll_area.setWidget(self.tb_widget)

        self.grid.addWidget(self.tb_scroll_area, 1, 0, 1, 2)

        # Slots, signals and proxies
        # ==================================================

        # Mouse tracking for the ROI
        # ==================================================
        self.mouse_tracking_toggle = False
        self.mouse_tracker = SignalProxy(
            self.iv.scene.sigMouseMoved,
            rateLimit=100,
            slot=self._mouse_coordinates,
        )

        self.event_handler = Canvas.EventHandler(self)
        self.installEventFilter(self.event_handler)
        self.event_handler.ctrl_signal.connect(self._toggle_mouse_tracking)
        self.event_handler.focus_signal.connect(self._focus_canvas)

    @property
    def layer(self) -> Layer:
        return self.stack.layers[self.stack.current_layer]

    def set_stack(
        self,
        stack: Stack,
        auto_range: bool = False,
    ):
        # Update the stack
        # ==================================================
        self.stack = stack
        if len(self.stack.layers) > 0:
            self.stack._update_current_layer()

        # Update the thumbnails
        # ==================================================
        self._update_thumbnails()

        # Select the first layer
        # ==================================================
        self.slot_select_layer(0, auto_range)

    def process(
        self,
        auto_range: bool = False,
    ):
        # Paint the result onto the canvas
        # ==================================================
        self._draw(auto_range)

        # Set up the ROI
        # ==================================================
        self.iv.set_roi(self.layer)

        # Plot the radial and phase profiles
        # ==================================================
        self.plot.emit()

    def _draw(
        self,
        auto_range: bool = False,
    ):
        # Coordinates of the current layer and the stack
        # ==================================================
        lcx, lcy = self.layer.centre
        scx, scy = self.stack.centre

        # # Reset the canvas
        # # ==================================================
        self.image = np.zeros(self.stack.merged.shape + (4,))

        # Draw the slice and potentially the stack
        # ==================================================
        # if conf.show_stack:
        #     idx = np.argwhere(self.stack.merged > 0).T
        #     self.image[idx[0], idx[1], 0] = conf.stack_contour_colour.red()
        #     self.image[idx[0], idx[1], 1] = conf.stack_contour_colour.green()
        #     self.image[idx[0], idx[1], 2] = conf.stack_contour_colour.blue()
        #     self.image[idx[0], idx[1], 3] = conf.stack_contour_colour.alpha()

        idx = np.argwhere(self.layer.image > 0).T
        self.image[idx[0], idx[1], :3] = ski.exposure.rescale_intensity(
            self.layer.image[idx[0], idx[1], None],
            in_range=np.ubyte,
            out_range=(0.0, 1.0),
        ) * np.array(
            [
                conf.slice_contour_colour.red(),
                conf.slice_contour_colour.green(),
                conf.slice_contour_colour.blue(),
            ]
        )
        # self.image[..., :3] = ski.exposure.rescale_intensity(
        #     self.layer.image[:,:,None],
        #     in_range=np.ubyte,
        #     out_range=(0.0, 1.0),
        # ) * np.array(
        #     [
        #         conf.slice_contour_colour.red(),
        #         conf.slice_contour_colour.green(),
        #         conf.slice_contour_colour.blue(),
        #     ]
        # )
        # self.image[idx[0], idx[1], 1] = self.layer.image[..., 1] # * conf.slice_contour_colour.green()
        # self.image[idx[0], idx[1], 2] = self.layer.image[..., 2] # * conf.slice_contour_colour.blue()
        self.image[idx[0], idx[1], 3] = conf.slice_contour_colour.alpha()

        # Draw the centre
        # ==================================================
        if self.slice_centre is not None:
            self.iv.removeItem(self.slice_centre)
        self.slice_centre = Target(
            (lcy + 0.5, lcx + 0.5),
            pen=self.slice_centre_pen,
        )
        self.iv.addItem(self.slice_centre)

        # Draw the slice contour
        # ==================================================
        if self.slice_contour is not None:
            self.iv.removeItem(self.slice_contour)
        self.slice_contour = Contour(
            (lcy + 0.5, lcx + 0.5),
            radius=self.layer.mbr,
            pen=self.slice_contour_pen,
        )
        self.iv.addItem(self.slice_contour)

        # Draw the stack contour
        # ==================================================
        if self.stack_contour is not None:
            self.iv.removeItem(self.stack_contour)
        self.stack_contour = Contour(
            (scy + 0.5, scx + 0.5),
            radius=self.stack.radius,
            pen=self.stack_contour_pen,
        )
        self.iv.addItem(self.stack_contour)

        # Set the image
        # ==================================================
        self.iv.setImage(
            self.image, autoRange=auto_range, levels=(0, 255), levelMode="rgba"
        )

    def _update_thumbnails(self):
        while True:
            item = self.tb_layout.takeAt(0)
            if item is None or item.isEmpty():
                break
            self.tb_layout.removeWidget(item.widget())
            item.widget().deleteLater()

        self.thumbnails.clear()
        for index, layer in enumerate(self.stack.layers):
            tb = Thumbnail(index, layer, self, 90)
            tb._selected.connect(self.slot_select_layer)
            self.thumbnails.append(tb)

        for idx, tb in enumerate(self.thumbnails):
            self.tb_layout.addWidget(tb, alignment=Qt.AlignmentFlag.AlignLeft)

        self.tb_layout.addStretch()

    @Slot()
    def _reset_zoom(self):
        self._draw(auto_range=True)

    def eventFilter(self, obj, event):
        if obj is self.window:
            if event.type() == QEvent.KeyPress:
                if event.key() == Qt.Key_Control:
                    self.ctrl_signal.emit(True)
            if event.type() == QEvent.KeyRelease:
                if event.key() == Qt.Key_Control:
                    self.ctrl_signal.emit(False)
        return super().eventFilter(obj, event)

    @Slot()
    def _mouse_coordinates(
        self,
        event: MouseClickEvent,
    ):

        if self.mouse_tracking_toggle:
            if self.iv.radial_roi is not None:
                pos = self.iv.getView().vb.mapSceneToView(event)[0]

                cx = pos.x() - self.layer.centre[1] - 0.5
                cy = pos.y() - self.layer.centre[0] - 0.5
                r = np.sqrt(cx**2 + cy**2)
                radius = r / self.layer.mbr
                r2 = 2 * r
                if radius <= 1:
                    phase = 0 if r == 0 else np.rad2deg(np.arccos(cx / r))

                    if cy < 0:
                        phase = 360 - phase

                    self.iv.radial_roi.setSize(
                        r2,
                        center=(0.5, 0.5),
                        update=True,
                        finish=True,
                    )
                    self.iv.phase_roi.set_end(pos)
                    self.update_radial_plot.emit(radius)
                    self.update_phase_plot.emit(phase)

    @Slot(bool)
    def _toggle_mouse_tracking(
        self,
        toggle: bool,
    ):
        self.mouse_tracking_toggle = toggle
        # self.iv.phase_roi.setVisible(toggle)
        # self.iv.radial_roi.setVisible(toggle)

    @Slot()
    def _focus_canvas(self):
        self.setFocus()

    @Slot()
    def _set_slice_contour_colour(self):
        self.slice_contour_pen = pg.mkPen(color=conf.slice_contour_colour, width=1)
        self._draw()

    @Slot()
    def _set_stack_contour_colour(self):
        self.stack_contour_pen = pg.mkPen(color=conf.stack_contour_colour, width=1)
        self._draw()

    @Slot()
    def _set_show_stack(self):
        self._draw()

    @Slot(int)
    def _slot_compute_radial_profile(self, segments: int):
        if self.layer is None:
            return

        self.layer.compute_radial_profile()
        self.plot.emit()

    @Slot(int)
    def _slot_compute_phase_profile(self, segments: int):
        if self.layer is None:
            return

        self.layer.compute_phase_profile()
        self.plot.emit()

    @Slot(int, bool)
    def slot_select_layer(
        self,
        layer: int,
        auto_range: bool = False,
    ):
        cur_layer = self.stack.current_layer
        if cur_layer is None:
            cur_layer = 0

        # Update the layer
        self.stack._update_current_layer(layer)

        # Highlight the selected thumbnail
        self.thumbnails[cur_layer].deselect()
        self.thumbnails[layer].select()

        # Process the layer
        self.process(auto_range)

        # Emit a signal to highlight the relevant plots
        self.highlight_plot.emit(layer)



import pyqtgraph as pg


class Target(pg.TargetItem):
    def __init__(
        self,
        pos: tuple,
        *args,
        **kwargs,
    ):

        kwargs.setdefault("movable", False)
        super().__init__(pos, *args, **kwargs)



from PySide6.QtCore import Qt
from PySide6.QtCore import Signal
from PySide6.QtCore import Slot

from PySide6.QtWidgets import QWidget
from PySide6.QtWidgets import QDockWidget
from PySide6.QtWidgets import QPushButton
from PySide6.QtWidgets import QCheckBox
from PySide6.QtWidgets import QSpinBox
from PySide6.QtWidgets import QLabel
from PySide6.QtWidgets import QFormLayout
from PySide6.QtWidgets import QSizePolicy

from mimetica import conf
from mimetica.utils.functions import as_rgba
from mimetica.utils.functions import get_colour


class Dock(QDockWidget):
    sig_show_inactive_plots = Signal()
    sig_set_active_plot_colour = Signal()
    sig_set_inactive_plot_colour = Signal()
    sig_set_stack_contour_colour = Signal()
    sig_set_slice_contour_colour = Signal()
    sig_show_stack = Signal()
    sig_set_radial_segments = Signal(int)
    sig_set_phase_segments = Signal(int)

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        title = QLabel("Settings", self)
        title.setAlignment(Qt.AlignmentFlag.AlignCenter)
        title.setStyleSheet("font-size: 14pt; font-weight: bold; text-align: center;")
        title.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)
        self.setTitleBarWidget(title)

        self.setVisible(False)
        self.setMinimumWidth(200)
        topright = Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignRight

        self.grid = QFormLayout()

        # Show all plots checkbox
        # ==================================================
        self.show_inactive_plots_lbl = QLabel(f"Show all plots:", self)
        self.show_inactive_plots_cbox = QCheckBox(self)
        self.show_inactive_plots_cbox.setChecked(conf.show_inactive_plots)
        self.grid.addRow(self.show_inactive_plots_lbl, self.show_inactive_plots_cbox)
        self.show_inactive_plots_cbox.stateChanged.connect(
            self._slot_show_inactive_plots
        )

        # Active plot colour
        # ==================================================
        self.active_plot_colour_lbl = QLabel(f"Active plot colour:", self)
        self.active_plot_colour_btn = QPushButton(self)
        self.active_plot_colour_btn.setStyleSheet(
            f"background-color:rgba({as_rgba(conf.active_plot_colour)});"
        )
        self.grid.addRow(self.active_plot_colour_lbl, self.active_plot_colour_btn)
        self.active_plot_colour_btn.pressed.connect(self._slot_set_active_plot_colour)

        # Inactive plot colour
        # ==================================================
        self.inactive_plot_colour_lbl = QLabel(f"Inactive plot colour:", self)
        self.inactive_plot_colour_btn = QPushButton(self)
        self.inactive_plot_colour_btn.setStyleSheet(
            f"background-color:rgba({as_rgba(conf.inactive_plot_colour)});"
        )
        self.grid.addRow(self.inactive_plot_colour_lbl, self.inactive_plot_colour_btn)
        self.inactive_plot_colour_btn.pressed.connect(
            self._slot_set_inactive_plot_colour
        )

        # Stack contour colour
        # ==================================================
        self.stack_contour_colour_lbl = QLabel(f"Stack contour colour:", self)
        self.stack_contour_colour_btn = QPushButton(self)
        self.stack_contour_colour_btn.setStyleSheet(
            f"background-color:rgba({as_rgba(conf.stack_contour_colour)});"
        )
        self.grid.addRow(self.stack_contour_colour_lbl, self.stack_contour_colour_btn)
        self.stack_contour_colour_btn.pressed.connect(
            self._slot_set_stack_contour_colour
        )

        # Slice contour colour
        # ==================================================
        self.slice_contour_colour_lbl = QLabel(f"Slice contour colour:", self)
        self.slice_contour_colour_btn = QPushButton(self)
        self.slice_contour_colour_btn.setStyleSheet(
            f"background-color:rgba({as_rgba(conf.slice_contour_colour)});"
        )
        self.grid.addRow(self.slice_contour_colour_lbl, self.slice_contour_colour_btn)
        self.slice_contour_colour_btn.pressed.connect(
            self._slot_set_slice_contour_colour
        )

        # Show the stack
        # ==================================================
        self.show_stack_lbl = QLabel(f"Show the stack:", self)
        self.show_stack_cbox = QCheckBox(self)
        self.show_stack_cbox.setChecked(conf.show_stack)
        self.grid.addRow(self.show_stack_lbl, self.show_stack_cbox)
        self.show_stack_cbox.stateChanged.connect(self._slot_show_stack)

        # Radial segments
        # ==================================================
        self.radial_segments_lbl = QLabel(f"Radial segments:", self)
        self.radial_segments_sbox = QSpinBox(
            self,
            value=conf.radial_segments,
            minimum=1,
            maximum=1024,
            singleStep=1,
        )
        self.radial_segments_sbox.setValue(conf.radial_segments)
        self.grid.addRow(self.radial_segments_lbl, self.radial_segments_sbox)
        self.radial_segments_sbox.valueChanged.connect(self._slot_set_radial_segments)

        # Phase segments
        # ==================================================
        self.phase_segments_lbl = QLabel(f"Phase segments:", self)
        self.phase_segments_sbox = QSpinBox(
            self,
            value=conf.phase_segments,
            minimum=120,
            maximum=720,
            singleStep=1,
            suffix="°",
        )
        self.phase_segments_sbox.setValue(conf.phase_segments)
        self.grid.addRow(self.phase_segments_lbl, self.phase_segments_sbox)
        self.phase_segments_sbox.valueChanged.connect(self._slot_set_phase_segments)

        # Set up the main widget
        # ==================================================
        self.body = QWidget()
        self.setWidget(self.body)
        self.body.setLayout(self.grid)
        self.body.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)

    @Slot()
    def _slot_show_inactive_plots(self):
        conf.show_inactive_plots = self.show_inactive_plots_cbox.isChecked()
        self.sig_show_inactive_plots.emit()

    @Slot()
    def _slot_set_active_plot_colour(self):
        colour = get_colour(conf.active_plot_colour, self)
        if colour.isValid():
            conf.active_plot_colour = colour
            self.sig_set_active_plot_colour.emit()
            self.active_plot_colour_btn.setStyleSheet(
                f"background-color:rgba({as_rgba(colour)});"
            )

    @Slot()
    def _slot_set_inactive_plot_colour(self):
        colour = get_colour(conf.inactive_plot_colour, self)
        if colour.isValid():
            conf.inactive_plot_colour = colour
            self.sig_set_inactive_plot_colour.emit()
            self.inactive_plot_colour_btn.setStyleSheet(
                f"background-color:rgba({as_rgba(colour)});"
            )

    @Slot()
    def _slot_set_stack_contour_colour(self):
        colour = get_colour(conf.stack_contour_colour, self)
        if colour.isValid():
            conf.stack_contour_colour = colour
            self.sig_set_stack_contour_colour.emit()
            self.stack_contour_colour_btn.setStyleSheet(
                f"background-color:rgba({as_rgba(colour)});"
            )

    @Slot()
    def _slot_set_slice_contour_colour(self):
        colour = get_colour(conf.slice_contour_colour, self)
        if colour.isValid():
            conf.slice_contour_colour = colour
            self.sig_set_slice_contour_colour.emit()
            self.slice_contour_colour_btn.setStyleSheet(
                f"background-color:rgba({as_rgba(colour)});"
            )

    @Slot()
    def _slot_show_stack(self):
        conf.show_stack = self.show_stack_cbox.isChecked()
        self.sig_show_stack.emit()

    @Slot()
    def _slot_set_radial_segments(self):
        conf.radial_segments = self.radial_segments_sbox.value()
        self.sig_set_radial_segments.emit(conf.radial_segments)

    @Slot()
    def _slot_set_phase_segments(self):
        conf.phase_segments = self.phase_segments_sbox.value()
        self.sig_set_phase_segments.emit(conf.phase_segments)




import numpy as np

from PySide6.QtCore import Qt
from PySide6.QtCore import Slot

from PySide6.QtWidgets import QSplitter

from PySide6.QtGui import QColor

import pyqtgraph as pg

from mimetica import conf
from mimetica import Canvas
from mimetica import Plot
from mimetica import Layer


class SplitView(QSplitter):

    def __init__(
        self,
        canvas: Canvas,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        # Canvas
        # ==================================================
        self.canvas = canvas
        self.addWidget(self.canvas)

        self.right_panel = QSplitter(self)
        self.right_panel.setOrientation(Qt.Orientation.Vertical)
        self.addWidget(self.right_panel)

        # Radial profile plot
        # ==================================================
        self.radial_graph = Plot(
            title="Radial profile",
            labels={
                "left": "Density [a.u.]",
                "bottom": "Normalised radius [a.u.]",
            },
        )
        self.radial_graph.setXRange(0, 1)
        self.radial_graph.setYRange(0, 1.2)
        self.radial_plots = {}
        self.right_panel.addWidget(self.radial_graph)
        self.radial_guide = pg.InfiniteLine(0)
        self.radial_arrow = None
        self.radial_lbl = pg.InfLineLabel(
            line=self.radial_guide,
            text="",
            position=0.98,
        )

        # Phase profile plot
        # ==================================================
        self.phase_graph = Plot(
            title="Phase profile",
            labels={
                "left": "Density [a.u.]",
                "bottom": "Angle [deg]",
            },
        )
        self.phase_graph.setXRange(0, 360)
        self.phase_graph.setYRange(0, 1.2)
        self.phase_plots = {}
        self.right_panel.addWidget(self.phase_graph)
        self.phase_arrow = None
        self.phase_guide = pg.InfiniteLine(0)
        self.phase_lbl = pg.InfLineLabel(
            line=self.phase_guide,
            text="",
            position=0.98,
        )

        # Pens
        # ==================================================
        # TODO: Migrate to conf
        self.active_pen_width = 1.0
        self.active_pen = pg.mkPen(
            color=conf.active_plot_colour,
            width=self.active_pen_width,
        )

        # TODO: Migrate to conf
        self.inactive_pen_width = 0.5
        self.inactive_pen = pg.mkPen(
            color=conf.inactive_plot_colour,
            width=self.inactive_pen_width,
        )
        self.invisible_pen = pg.mkPen(color=QColor(0, 0, 0, 0), width=0)

        # Current layer index & reference
        # ==================================================
        self.current_layer_idx = 0
        self.current_layer = None

        # Slots and signals
        # ==================================================
        self.canvas.update_radial_plot.connect(self._update_radial_pos)
        self.canvas.update_phase_plot.connect(self._update_phase_pos)
        self.canvas.plot.connect(
            lambda: self.plot(self.canvas.stack.layers, self.canvas.stack.current_layer)
        )
        self.canvas.highlight_plot.connect(lambda layer: self._highlight_plot(layer))

    @Slot(list, int)
    def plot(
        self,
        layers: list[Layer],
        current_layer_idx: int,
    ):
        self.radial_plots.clear()
        self.phase_plots.clear()

        self.radial_graph.clear()
        self.phase_graph.clear()

        for radial_plot in self.radial_plots.values():
            self.radial_graph.removeItem(radial_plot)

        for phase_plot in self.phase_plots.values():
            self.phase_graph.removeItem(phase_plot)

        # Plot all layers
        # ==================================================
        for idx, layer in enumerate(layers):
            if idx == current_layer_idx:
                pen = self.active_pen
                self.current_layer_idx = idx
                self.current_layer = layer
                self.phase_graph.setXRange(0.0, layer.phase_range.max())
            else:

                pen = (
                    self.inactive_pen
                    if conf.show_inactive_plots
                    else self.invisible_pen
                )

            # Plot the radial profile
            if layer.radial_range is not None:
                self.radial_plots[idx] = self.radial_graph.plot(
                    layer.radial_range,
                    layer.radial_profile,
                    pen=pen,
                )

            # Plot the phase profile
            if layer.phase_range is not None:
                self.phase_plots[idx] = self.phase_graph.plot(
                    layer.phase_range,
                    layer.phase_profile,
                    pen=pen,
                )

            if idx == current_layer_idx:
                self.radial_arrow = pg.CurveArrow(self.radial_plots[idx])
                self.radial_arrow.setRotation(270)
                self.radial_arrow._rotate = False
                self.phase_arrow = pg.CurveArrow(self.phase_plots[idx])
                self.phase_arrow.setRotation(270)
                self.phase_arrow._rotate = False

        # Add vertical guides
        # ==================================================
        self.radial_graph.addItem(self.radial_guide)
        self.radial_graph.addItem(self.radial_arrow)
        self.phase_graph.addItem(self.phase_guide)
        self.phase_graph.addItem(self.phase_arrow)

    @Slot(float, str)
    def _update_radial_pos(
        self,
        radius: float,
    ):
        # Find the index of the closest value in self.radial_xs
        xs = self.current_layer.radial_range
        diff = np.absolute(xs - radius)
        index = diff.argmin()
        x = xs[index]
        y = self.current_layer.radial_profile[index]
        self.radial_guide.setPos(x)
        self.radial_arrow.setIndex(int(index))
        self.radial_lbl.setText(f"{x.item():0.3f}|{y:0.3f}")

    @Slot(float, str)
    def _update_phase_pos(
        self,
        phase: float,
    ):
        # Find the index of the closest phase value
        xs = self.current_layer.phase_range
        diff = np.absolute(xs - phase)
        index = diff.argmin()
        x = xs[index]
        y = self.current_layer.phase_profile[index]
        self.phase_guide.setPos(x)
        self.phase_arrow.setIndex(index)
        self.phase_lbl.setText(f"{x.item():3.3f}|{y:0.3f}")

    @Slot()
    def _show_inactive_plots(self):

        for idx in self.radial_plots:
            if idx != self.current_layer_idx:
                if conf.show_inactive_plots:
                    self.radial_plots[idx].setPen(self.inactive_pen)
                    self.phase_plots[idx].setPen(self.inactive_pen)
                    self.radial_plots[idx].update()
                    self.phase_plots[idx].update()
                else:
                    self.radial_plots[idx].setPen(self.invisible_pen)
                    self.phase_plots[idx].setPen(self.invisible_pen)
                    self.radial_plots[idx].update()
                    self.phase_plots[idx].update()

    @Slot()
    def _set_inactive_plot_colour(self):

        self.inactive_pen = pg.mkPen(
            color=conf.inactive_plot_colour,
            width=self.inactive_pen_width,
        )
        if conf.show_inactive_plots:
            for idx in self.radial_plots:
                if idx != self.current_layer_idx:
                    self.radial_plots[idx].setPen(self.inactive_pen)
                    self.radial_plots[idx].update()
                    self.phase_plots[idx].setPen(self.inactive_pen)
                    self.phase_plots[idx].update()

    @Slot()
    def _set_active_plot_colour(self):

        self.active_pen = pg.mkPen(
            color=conf.active_plot_colour,
            width=self.active_pen_width,
        )
        idx = self.current_layer_idx
        self.radial_plots[idx].setPen(self.active_pen)
        self.radial_plots[idx].update()
        self.phase_plots[idx].setPen(self.active_pen)
        self.phase_plots[idx].update()

    @Slot(int)
    def _highlight_plot(
        self,
        layer_idx: int,
    ):

        self.radial_plots[self.current_layer_idx].setPen(self.inactive_pen)
        self.radial_plots[layer_idx].setPen(self.active_pen)
        self.radial_plots[layer_idx].update()
        self.radial_graph.getPlotItem().enableAutoRange()

        self.phase_plots[self.current_layer_idx].setPen(self.inactive_pen)
        self.phase_plots[layer_idx].setPen(self.active_pen)
        self.phase_plots[layer_idx].update()
        self.phase_graph.getPlotItem().enableAutoRange()

        self.current_layer_idx = layer_idx



import skimage as ski
import shapely as shp
from shapely.geometry import Polygon
import numpy as np
from PySide6.QtGui import QColor
from PySide6.QtCore import QObject
from PySide6.QtWidgets import QColorDialog


def compute_minimal_bounding_circle(image: np.ndarray):
    '''
    Compute the minimal bounding circle (MBC) and radius.

    Args:
        image: The image being analysed.

    Returns:
        The MBC and its radius.
    '''

    point_set = np.argwhere(image)
    points = shp.MultiPoint(point_set)

    mbc = shp.minimum_bounding_circle(points)
    mbr = shp.minimum_bounding_radius(points)
    return mbc, mbr


def draw_sorted_circle(centre: np.ndarray, radius: np.ndarray):
    rr, cc = ski.draw.circle_perimeter(
        *centre.astype(np.uint32),
        np.floor(radius).astype(np.uint32),
        method="andres",
    )
    rr = np.array(rr, dtype=np.uint32)
    cc = np.array(cc, dtype=np.uint32)
    angle = np.argsort(np.arctan2(cc - np.mean(cc), rr - np.mean(rr)))
    rr = np.take(rr, angle, axis=0)
    cc = np.take(cc, angle, axis=0)
    return rr, cc


def as_rgba(colour: QColor) -> str:
    """
    Convert a QColor isntance (including opacity) into RGBA format
    suitable for use in a stylesheet.

    The R, G and B values are between 0 and 255, and the opacity
    is between 0 and 1.

    Args:
        colour:
            A QColor instance.

    Returns:
        RGBA-formatted string.
    """
    return ",".join(
        str(c)
        for c in [
            colour.red(),
            colour.green(),
            colour.blue(),
            colour.alphaF(),
        ]
    )


def as_hex(colour: QColor) -> str:
    """
    Convert a QColor isntance (including opacity) into hex format.

    Args:
        colour:
            A QColor instance.

    Returns:
        The hex representation.
    """
    return "#" + "".join(
        [
            f"{colour.red():02X}",
            f"{colour.green():02X}",
            f"{colour.blue():02X}",
            f"{colour.alpha():02X}",
        ]
    )


def get_colour(
    initial: QColor = None,
    parent: QObject = None,
    title: str = "Pick a colour",
) -> QColor:
    """
    Get a QColor from a QColorDialog.

    Returns:
        The selected colour.
    """

    return QColorDialog.getColor(
        initial,
        parent,
        title,
        QColorDialog.ColorDialogOption.ShowAlphaChannel,
    )
